Public Class NodeSearch(Of T)

    Public Property Item As T
    Public Property IsVisited As Boolean?

    Public Property Neighbors As New List(Of NodeSearch(Of T))
    Public Property Weights As New List(Of Double?)

    Public Overrides Function ToString() As String
        Return If(Item?.ToString(), "")
    End Function

End Class


Public Class GraphSearch(Of T)
    Private directed As Boolean

    Public Sub New(Optional directed As Boolean = False)
        Me.directed = directed
    End Sub

    Public Property Nodes As List(Of NodeSearch(Of T)) = New List(Of NodeSearch(Of T))

    Public Function AddNode(item As T) As NodeSearch(Of T)
        Dim node As New NodeSearch(Of T) With {.Item = item}
        Nodes.Add(node)
        Return node
    End Function

    Public Sub AddEdge(fromNode As NodeSearch(Of T), toNode As NodeSearch(Of T), Optional weight As Double? = Nothing)
        fromNode.Neighbors.Add(toNode)
        fromNode.Weights.Add(weight)

        If Not directed Then
            toNode.Neighbors.Add(fromNode)
            toNode.Weights.Add(weight)
        End If
    End Sub

    Default Public ReadOnly Property Item(itemData As T) As NodeSearch(Of T)
        Get
            Dim node = Nodes.Find(Function(n) n.Item?.Equals(itemData) = True)

            If node IsNot Nothing Then
                Return node
            End If

            Throw New IndexOutOfRangeException("Item not found")
        End Get
    End Property
    Public Function Contains(item As T) As Boolean
        Dim node = Nodes.Find(Function(n) n.Item?.Equals(item) = True)
        Return node IsNot Nothing
    End Function

    Public Function DFS() As List(Of NodeSearch(Of T))
        Dim visited As New List(Of NodeSearch(Of T))

        Nodes.ForEach(Sub(n) n.IsVisited = Nothing)

        Dim stack As New Stack(Of NodeSearch(Of T))

        Nodes(0).IsVisited = True
        stack.Push(Nodes(0))
        visited.Add(Nodes(0))

        While stack.Count > 0
            Dim neighbor = GetNextUnvisitedNeighbor(stack.Peek())

            If neighbor Is Nothing Then
                stack.Pop()
            Else
                neighbor.IsVisited = True
                visited.Add(neighbor)
                stack.Push(neighbor)
            End If
        End While

        Nodes.ForEach(Sub(n) n.IsVisited = Nothing)

        Return visited
    End Function

    Private Shared Function GetNextUnvisitedNeighbor(node As NodeSearch(Of T)) As NodeSearch(Of T)
        For Each neighbor In node.Neighbors
            If neighbor.IsVisited Is Nothing Then
                Return neighbor
            End If
        Next
        Return Nothing
    End Function

    Public Function BFS() As List(Of NodeSearch(Of T))
        Dim visited As New List(Of NodeSearch(Of T))

        Nodes.ForEach(Sub(n) n.IsVisited = Nothing)

        Dim queue As New Queue(Of NodeSearch(Of T))

        Nodes(0).IsVisited = True
        queue.Enqueue(Nodes(0))

        While queue.Count > 0
            Dim node = queue.Dequeue()
            visited.Add(node)

            For Each neighbor In node.Neighbors
                If neighbor.IsVisited Is Nothing Then
                    neighbor.IsVisited = True
                    queue.Enqueue(neighbor)
                End If
            Next
        End While

        Nodes.ForEach(Sub(n) n.IsVisited = Nothing)

        Return visited
    End Function

    Public Function DFSRecursive() As List(Of NodeSearch(Of T))
        Dim list = New List(Of NodeSearch(Of T))()

        Nodes.ForEach(Sub(n As Object) n.IsVisited = Nothing)
        DFSRecursive(Nodes(0), list)
        Nodes.ForEach(Sub(n As Object) n.IsVisited = Nothing)

        Return list
    End Function

    Private Sub DFSRecursive(node As NodeSearch(Of T), list As List(Of NodeSearch(Of T)))
        list.Add(node)
        node.IsVisited = True

        For Each neighbor In node.Neighbors
            If neighbor.IsVisited Is Nothing Then
                DFSRecursive(neighbor, list)
            End If
        Next
    End Sub

    Public Overrides Function ToString() As String
        Dim s = ""

        Nodes.ForEach(Sub(node)
                          s += $"{node} -> "

                          Dim neighbors As New List(Of String)()
                          For i = 0 To node.Neighbors.Count - 1
                              Dim neighbor = node.Neighbors(i)
                              Dim weight = node.Weights(i)
                              neighbors.Add($"{neighbor} ({weight})")
                          Next
                          s += String.Join(", ", neighbors) + Environment.NewLine
                      End Sub)
        Return s
    End Function

End Class

